@@ -45,8 +45,6 @@ module Development.IDE.GHC.Compat(
     readHieFile,
     setHieDir,
     dontWriteHieFiles,
-    module Compat.HieTypes,
-    module Compat.HieUtils,
     -- * Compat modules
     module Development.IDE.GHC.Compat.Core,
     module Development.IDE.GHC.Compat.Env,
@@ -68,6 +66,7 @@ module Development.IDE.GHC.Compat(
     simplifyExpr,
     tidyExpr,
     emptyTidyEnv,
+    tidyOpenType,
     corePrepExpr,
     corePrepPgm,
     lintInteractiveExpr,
@@ -75,6 +74,7 @@ module Development.IDE.GHC.Compat(
     HomePackageTable,
     lookupHpt,
     loadModulesHome,
+    hugElts,
     bcoFreeNames,
     ModIfaceAnnotation,
     pattern Annotation,
@@ -89,7 +89,9 @@ module Development.IDE.GHC.Compat(
     emptyInScopeSet,
     Unfolding(..),
     noUnfolding,
+#if !MIN_VERSION_ghc(9,13,0)
     loadExpr,
+#endif
     byteCodeGen,
     bc_bcos,
     loadDecls,
@@ -102,9 +104,7 @@ module Development.IDE.GHC.Compat(
     Dependencies(dep_direct_mods),
     NameCacheUpdater,
 
-#if MIN_VERSION_ghc(9,5,0)
     XModulePs(..),
-#endif
 
 #if !MIN_VERSION_ghc(9,7,0)
     liftZonkM,
@@ -114,14 +114,8 @@ module Development.IDE.GHC.Compat(
 #if MIN_VERSION_ghc(9,7,0)
     tcInitTidyEnv,
 #endif
-    ) where
 
-import           Compat.HieAst                           (enrichHie)
-import           Compat.HieBin
-import           Compat.HieTypes                         hiding
-                                                         (nodeAnnotations)
-import qualified Compat.HieTypes                         as GHC (nodeAnnotations)
-import           Compat.HieUtils
+    ) where
 import           Control.Applicative                     ((<|>))
 import qualified Data.ByteString                         as BS
 import           Data.Coerce                             (coerce)
@@ -145,15 +139,22 @@ import           Prelude                                 hiding (mod)
 
 import qualified GHC.Core.Opt.Pipeline                   as GHC
 import           GHC.Core.Tidy                           (tidyExpr)
+import           GHC.Core.TyCo.Tidy                      (tidyOpenType)
 import           GHC.CoreToStg.Prep                      (corePrepPgm)
 import qualified GHC.CoreToStg.Prep                      as GHC
 import           GHC.Driver.Hooks                        (hscCompileCoreExprHook)
+import           GHC.Iface.Ext.Types                     hiding
+                                                         (nodeAnnotations)
+import qualified GHC.Iface.Ext.Types                     as GHC (nodeAnnotations)
+import           GHC.Iface.Ext.Utils
 
 import           GHC.ByteCode.Asm                        (bcoFreeNames)
 import           GHC.Core
 import           GHC.Data.FastString
 import           GHC.Data.StringBuffer
 import           GHC.Driver.Session                      hiding (ExposePackage)
+import           GHC.Iface.Ext.Ast                       (enrichHie)
+import           GHC.Iface.Ext.Binary
 import           GHC.Iface.Make                          (mkIfaceExports)
 import           GHC.SysTools.Tasks                      (runPp, runUnlit)
 import           GHC.Types.Annotations                   (AnnTarget (ModuleTarget),
@@ -167,12 +168,21 @@ import           GHC.Types.Var.Env
 
 import           GHC.Builtin.Uniques
 import           GHC.ByteCode.Types
+import           GHC.Core.Lint.Interactive               (interactiveInScope)
 import           GHC.CoreToStg
 import           GHC.Data.Maybe
+import           GHC.Driver.Config.Core.Lint.Interactive (lintInteractiveExpr)
+import           GHC.Driver.Config.Core.Opt.Simplify     (initSimplifyExprOpts)
+import           GHC.Driver.Config.CoreToStg             (initCoreToStgOpts)
+import           GHC.Driver.Config.CoreToStg.Prep        (initCorePrepConfig)
 import           GHC.Driver.Config.Stg.Pipeline
 import           GHC.Driver.Env                          as Env
 import           GHC.Iface.Env
-import           GHC.Linker.Loader                       (loadDecls, loadExpr)
+import           GHC.Linker.Loader                       (loadDecls
+#if !MIN_VERSION_ghc(9,13,0)
+                                                         , loadExpr
+#endif
+                                                         )
 import           GHC.Runtime.Context                     (icInteractiveModule)
 import           GHC.Stg.Pipeline
 import           GHC.Stg.Syntax
@@ -180,24 +190,21 @@ import           GHC.StgToByteCode
 import           GHC.Types.CostCentre
 import           GHC.Types.IPE
 import           GHC.Types.SrcLoc                        (combineRealSrcSpans)
+#if MIN_VERSION_ghc(9,13,0)
+import           GHC.Unit.Home.PackageTable              (HomePackageTable,
+                                                          lookupHpt)
+#else
 import           GHC.Unit.Home.ModInfo                   (HomePackageTable,
                                                           lookupHpt)
+#endif
 import           GHC.Unit.Module.Deps                    (Dependencies (dep_direct_mods),
                                                           Usage (..))
 import           GHC.Unit.Module.ModIface
 
 -- See Note [Guidelines For Using CPP In GHCIDE Import Statements]
 
-#if !MIN_VERSION_ghc(9,5,0)
-import           GHC.Core.Lint                           (lintInteractiveExpr)
-#endif
-
-#if MIN_VERSION_ghc(9,5,0)
-import           GHC.Core.Lint.Interactive               (interactiveInScope)
-import           GHC.Driver.Config.Core.Lint.Interactive (lintInteractiveExpr)
-import           GHC.Driver.Config.Core.Opt.Simplify     (initSimplifyExprOpts)
-import           GHC.Driver.Config.CoreToStg             (initCoreToStgOpts)
-import           GHC.Driver.Config.CoreToStg.Prep        (initCorePrepConfig)
+#if MIN_VERSION_ghc(9,13,0)
+import           GHC.Unit.Home.Graph                     (addHomeModInfoToHug, unitEnv_assocs)
 #endif
 
 #if MIN_VERSION_ghc(9,7,0)
@@ -230,11 +237,7 @@ myCoreToStgExpr logger dflags ictxt
        binding for the stg2stg step) -}
     let bco_tmp_id = mkSysLocal (fsLit "BCO_toplevel")
                                 (mkPseudoUniqueE 0)
-#if MIN_VERSION_ghc(9,5,0)
                                 ManyTy
-#else
-                                Many
-#endif
                                 (exprType prepd_expr)
     (stg_binds, prov_map, collected_ccs) <-
        myCoreToStg logger
@@ -258,27 +261,17 @@ myCoreToStg logger dflags ictxt
     let (stg_binds, denv, cost_centre_info)
          = {-# SCC "Core2Stg" #-}
            coreToStg
-#if MIN_VERSION_ghc(9,5,0)
              (initCoreToStgOpts dflags)
-#else
-             dflags
-#endif
              this_mod ml prepd_binds
 
 #if MIN_VERSION_ghc(9,8,0)
     (unzip -> (stg_binds2,_),_)
-#elif MIN_VERSION_ghc(9,4,2)
-    (stg_binds2,_)
 #else
-    stg_binds2
+    (stg_binds2,_)
 #endif
         <- {-# SCC "Stg2Stg" #-}
            stg2stg logger
-#if MIN_VERSION_ghc(9,5,0)
                    (interactiveInScope ictxt)
-#else
-                   ictxt
-#endif
                    (initStgPipelineOpts dflags for_bytecode) this_mod stg_binds
 
     return (stg_binds2, denv, cost_centre_info)
@@ -290,45 +283,28 @@ reLocA = reLoc
 #endif
 
 getDependentMods :: ModIface -> [ModuleName]
-getDependentMods = map (gwib_mod . snd) . S.toList . dep_direct_mods . mi_deps
-
-simplifyExpr :: DynFlags -> HscEnv -> CoreExpr -> IO CoreExpr
-#if MIN_VERSION_ghc(9,5,0)
-simplifyExpr _ env = GHC.simplifyExpr (Development.IDE.GHC.Compat.Env.hsc_logger env) (ue_eps (Development.IDE.GHC.Compat.Env.hsc_unit_env env)) (initSimplifyExprOpts (hsc_dflags env) (hsc_IC env))
+#if MIN_VERSION_ghc(9,13,0)
+getDependentMods = map (gwib_mod . (\(_,_,x) -> x)) . S.toList . dep_direct_mods . mi_deps
 #else
-simplifyExpr _ = GHC.simplifyExpr
+getDependentMods = map (gwib_mod . snd) . S.toList . dep_direct_mods . mi_deps
 #endif
 
+simplifyExpr :: DynFlags -> HscEnv -> CoreExpr -> IO CoreExpr
+simplifyExpr _ env = GHC.simplifyExpr (Development.IDE.GHC.Compat.Env.hsc_logger env) (ue_eps (Development.IDE.GHC.Compat.Env.hsc_unit_env env)) (initSimplifyExprOpts (hsc_dflags env) (hsc_IC env))
+
 corePrepExpr :: DynFlags -> HscEnv -> CoreExpr -> IO CoreExpr
-#if MIN_VERSION_ghc(9,5,0)
 corePrepExpr _ env expr = do
   cfg <- initCorePrepConfig env
   GHC.corePrepExpr (Development.IDE.GHC.Compat.Env.hsc_logger env) cfg expr
-#else
-corePrepExpr _ = GHC.corePrepExpr
-#endif
 
 renderMessages :: PsMessages -> (Bag WarnMsg, Bag ErrMsg)
 renderMessages msgs =
-#if MIN_VERSION_ghc(9,5,0)
   let renderMsgs extractor = (fmap . fmap) GhcPsMessage . getMessages $ extractor msgs
   in (renderMsgs psWarnings, renderMsgs psErrors)
-#else
-  let renderMsgs extractor = (fmap . fmap) renderDiagnosticMessageWithHints . getMessages $ extractor msgs
-  in (renderMsgs psWarnings, renderMsgs psErrors)
-#endif
 
-#if MIN_VERSION_ghc(9,5,0)
 pattern PFailedWithErrorMessages :: forall a b. (b -> Bag (MsgEnvelope GhcMessage)) -> ParseResult a
-#else
-pattern PFailedWithErrorMessages :: forall a b. (b -> Bag (MsgEnvelope DecoratedSDoc)) -> ParseResult a
-#endif
 pattern PFailedWithErrorMessages msgs
-#if MIN_VERSION_ghc(9,5,0)
      <- PFailed (const . fmap (fmap GhcPsMessage) . getMessages . getPsErrorMessages -> msgs)
-#else
-     <- PFailed (const . fmap (fmap renderDiagnosticMessageWithHints) . getMessages . getPsErrorMessages -> msgs)
-#endif
 {-# COMPLETE POk, PFailedWithErrorMessages #-}
 
 hieExportNames :: HieFile -> [(SrcSpan, Name)]
@@ -406,7 +382,11 @@ nameListFromAvails as =
 
 
 getModuleHash :: ModIface -> Fingerprint
+#if MIN_VERSION_ghc(9,13,0)
+getModuleHash = mi_mod_hash
+#else
 getModuleHash = mi_mod_hash . mi_final_exts
+#endif
 
 
 disableWarningsAsErrors :: DynFlags -> DynFlags
@@ -453,27 +433,27 @@ generatedNodeInfo :: HieAST a -> Maybe (NodeInfo a)
 generatedNodeInfo = Map.lookup GeneratedInfo . getSourcedNodeInfo . sourcedNodeInfo
 
 data GhcVersion
-  = GHC94
-  | GHC96
+  = GHC96
   | GHC98
   | GHC910
   | GHC912
+  | GHC914
   deriving (Eq, Ord, Show, Enum)
 
 ghcVersionStr :: String
 ghcVersionStr = VERSION_ghc
 
 ghcVersion :: GhcVersion
-#if MIN_VERSION_GLASGOW_HASKELL(9,12,0,0)
+#if MIN_VERSION_GLASGOW_HASKELL(9,14,0,0)
+ghcVersion = GHC914
+#elif MIN_VERSION_GLASGOW_HASKELL(9,12,0,0)
 ghcVersion = GHC912
 #elif MIN_VERSION_GLASGOW_HASKELL(9,10,0,0)
 ghcVersion = GHC910
 #elif MIN_VERSION_GLASGOW_HASKELL(9,8,0,0)
 ghcVersion = GHC98
-#elif MIN_VERSION_GLASGOW_HASKELL(9,6,0,0)
+#else
 ghcVersion = GHC96
-#elif MIN_VERSION_GLASGOW_HASKELL(9,4,0,0)
-ghcVersion = GHC94
 #endif
 
 simpleNodeInfoCompat :: FastStringCompat -> FastStringCompat -> NodeInfo a
@@ -504,20 +484,27 @@ mkAstNode n = Node (SourcedNodeInfo $ Map.singleton GeneratedInfo n)
 loadModulesHome
     :: [HomeModInfo]
     -> HscEnv
+#if MIN_VERSION_ghc(9,13,0)
+    -> IO HscEnv
+loadModulesHome mod_infos e = do
+  let hug = hsc_HUG (e { hsc_type_env_vars = emptyKnotVars })
+  mapM_ (`addHomeModInfoToHug` hug) mod_infos
+  pure (e { hsc_type_env_vars = emptyKnotVars })
+#else
     -> HscEnv
 loadModulesHome mod_infos e =
   hscUpdateHUG (\hug -> foldl' (flip addHomeModInfoToHug) hug mod_infos) (e { hsc_type_env_vars = emptyKnotVars })
+#endif
+
+#if MIN_VERSION_ghc(9,13,0)
+hugElts :: HomeUnitGraph -> [(UnitId, HomeUnitEnv)]
+hugElts = unitEnv_assocs
+#endif
 
 recDotDot :: HsRecFields (GhcPass p) arg -> Maybe Int
 recDotDot x =
-#if MIN_VERSION_ghc(9,5,0)
             unRecFieldsDotDot <$>
-#endif
             unLoc <$> rec_dotdot x
 
-#if MIN_VERSION_ghc(9,5,0)
-extract_cons (NewTypeCon x) = [x]
+extract_cons (NewTypeCon x)      = [x]
 extract_cons (DataTypeCons _ xs) = xs
-#else
-extract_cons = id
-#endif
