@@ -72,7 +72,9 @@ module Development.IDE.GHC.Compat.Core (
 #if MIN_VERSION_ghc(9,11,0)
     pattern ModIface,
     set_mi_top_env,
+#if !MIN_VERSION_ghc(9,13,0)
     set_mi_usages,
+#endif
 #endif
     HscSource(..),
     WhereFrom(..),
@@ -225,6 +227,7 @@ module Development.IDE.GHC.Compat.Core (
     SrcLoc.noSrcSpan,
     SrcLoc.noSrcLoc,
     SrcLoc.noLoc,
+    SrcLoc.srcSpanToRealSrcSpan,
     mapLoc,
     -- * Finder
     FindResult(..),
@@ -374,27 +377,13 @@ module Development.IDE.GHC.Compat.Core (
     module GHC.Unit.Finder.Types,
     module GHC.Unit.Env,
     module GHC.Driver.Phases,
-#if !MIN_VERSION_ghc(9,4,0)
-    pattern HsFieldBind,
-    hfbAnn,
-    hfbLHS,
-    hfbRHS,
-    hfbPun,
-#endif
-#if !MIN_VERSION_ghc_boot_th(9,4,1)
-    Extension(.., NamedFieldPuns),
-#else
     Extension(..),
-#endif
     mkCgInteractiveGuts,
     justBytecode,
     justObjects,
     emptyHomeModInfoLinkable,
     homeModInfoByteCode,
     homeModInfoObject,
-#if !MIN_VERSION_ghc(9,5,0)
-    field_label,
-#endif
     groupOrigin,
     isVisibleFunArg,
 #if MIN_VERSION_ghc(9,8,0)
@@ -535,6 +524,9 @@ import           GHC.Unit.Finder             hiding (mkHomeModLocation)
 import qualified GHC.Unit.Finder             as GHC
 import           GHC.Unit.Finder.Types
 import           GHC.Unit.Home.ModInfo
+#if MIN_VERSION_ghc(9,13,0)
+import           GHC.Unit.Home.PackageTable  (addToHpt, addListToHpt)
+#endif
 import           GHC.Unit.Module.Graph
 import           GHC.Unit.Module.Imported
 import           GHC.Unit.Module.ModDetails
@@ -547,7 +539,9 @@ import           GHC.Unit.Module.ModIface    (IfaceExport, ModIface,
 #if MIN_VERSION_ghc(9,11,0)
                                              , pattern ModIface
                                              , set_mi_top_env
+#if !MIN_VERSION_ghc(9,13,0)
                                              , set_mi_usages
+#endif
 #endif
                                              )
 import           GHC.Unit.Module.ModSummary  (ModSummary (..))
@@ -563,6 +557,10 @@ import           Language.Haskell.Syntax     hiding (FunDep)
 import System.OsPath
 #endif
 
+#if MIN_VERSION_ghc(9,13,0)
+import qualified System.FilePath as FP
+#endif
+
 #if !MIN_VERSION_ghc(9,7,0)
 import           GHC.Types.Avail             (greNamePrintableName)
 #endif
@@ -572,7 +570,17 @@ import           GHC.Hs                      (SrcSpanAnn')
 #endif
 
 mkHomeModLocation :: DynFlags -> ModuleName -> FilePath -> IO Module.ModLocation
-#if MIN_VERSION_ghc(9,11,0)
+#if MIN_VERSION_ghc(9,13,0)
+mkHomeModLocation df mn f =
+  let (basename, ext) = FP.splitExtension f
+      osBasename = unsafeEncodeUtf basename
+      osExt = unsafeEncodeUtf ext
+      hscSrc = case ext of
+        ".hs-boot" -> HsBootFile
+        ".hsig" -> HsigFile
+        _ -> HsSrcFile
+  in pure $ GHC.mkHomeModLocation (GHC.initFinderOpts df) mn osBasename osExt hscSrc
+#elif MIN_VERSION_ghc(9,11,0)
 mkHomeModLocation df mn f =
   let osf = unsafeEncodeUtf f
   in pure $ GHC.mkHomeModLocation (GHC.initFinderOpts df) mn osf
@@ -629,21 +637,11 @@ pattern ExposePackage s a mr = DynFlags.ExposePackage s a mr
 #endif
 
 isVisibleFunArg :: Development.IDE.GHC.Compat.Core.FunTyFlag -> Bool
-#if __GLASGOW_HASKELL__ >= 906
 isVisibleFunArg = TypesVar.isVisibleFunArg
 type FunTyFlag = TypesVar.FunTyFlag
-#else
-isVisibleFunArg VisArg = True
-isVisibleFunArg _ = False
-type FunTyFlag = TypesVar.AnonArgFlag
-#endif
 pattern FunTy :: Development.IDE.GHC.Compat.Core.FunTyFlag -> Type -> Type -> Type
 pattern FunTy af arg res <- TyCoRep.FunTy {ft_af = af, ft_arg = arg, ft_res = res}
 
-
--- type HasSrcSpan x a = (GenLocated SrcSpan a ~ x)
--- type HasSrcSpan x = () :: Constraint
-
 class HasSrcSpan a where
   getLoc :: a -> SrcSpan
 
@@ -702,7 +700,7 @@ initObjLinker env =
 loadDLL :: HscEnv -> String -> IO (Maybe String)
 loadDLL env str = do
     res <- GHCi.loadDLL (GHCi.hscInterp env) str
-#if MIN_VERSION_ghc(9,11,0) || (MIN_VERSION_ghc(9, 8, 3) && !MIN_VERSION_ghc(9, 9, 0))
+#if MIN_VERSION_ghc(9,11,0) || (MIN_VERSION_ghc(9, 8, 3) && !MIN_VERSION_ghc(9, 9, 0)) || (MIN_VERSION_ghc(9, 10, 2) && !MIN_VERSION_ghc(9, 11, 0))
     pure $
       case res of
         Left err_msg -> Just err_msg
@@ -749,11 +747,7 @@ makeSimpleDetails hsc_env =
 
 mkIfaceTc :: HscEnv -> GHC.SafeHaskellMode -> ModDetails -> ModSummary -> Maybe CoreProgram -> TcGblEnv -> IO ModIface
 mkIfaceTc hscEnv shm md _ms _mcp =
-#if MIN_VERSION_ghc(9,5,0)
   GHC.mkIfaceTc hscEnv shm md _ms _mcp -- mcp::Maybe CoreProgram is only used in GHC >= 9.6
-#else
-  GHC.mkIfaceTc hscEnv shm md _ms -- ms::ModSummary is only used in GHC >= 9.4
-#endif
 
 mkBootModDetailsTc :: HscEnv -> TcGblEnv -> IO ModDetails
 mkBootModDetailsTc session = GHC.mkBootModDetailsTc
@@ -767,50 +761,10 @@ initTidyOpts =
 driverNoStop :: StopPhase
 driverNoStop = NoStop
 
-
-#if !MIN_VERSION_ghc(9,4,0)
-pattern HsFieldBind :: XHsRecField id -> id -> arg -> Bool -> HsRecField' id arg
-pattern HsFieldBind {hfbAnn, hfbLHS, hfbRHS, hfbPun} <- HsRecField hfbAnn (SrcLoc.unLoc -> hfbLHS) hfbRHS hfbPun where
-  HsFieldBind ann lhs rhs pun = HsRecField ann (SrcLoc.noLoc lhs) rhs pun
-#endif
-
-#if !MIN_VERSION_ghc_boot_th(9,4,1)
-pattern NamedFieldPuns :: Extension
-pattern NamedFieldPuns = RecordPuns
-#endif
-
 groupOrigin :: MatchGroup GhcRn body -> Origin
-#if MIN_VERSION_ghc(9,5,0)
 mapLoc :: (a -> b) -> SrcLoc.GenLocated l a -> SrcLoc.GenLocated l b
 mapLoc = fmap
 groupOrigin = mg_ext
-#else
-mapLoc :: (a -> b) -> SrcLoc.GenLocated l a -> SrcLoc.GenLocated l b
-mapLoc = SrcLoc.mapLoc
-groupOrigin = mg_origin
-#endif
-
-
-#if !MIN_VERSION_ghc(9,5,0)
-mkCgInteractiveGuts :: CgGuts -> CgGuts
-mkCgInteractiveGuts = id
-
-emptyHomeModInfoLinkable :: Maybe Linkable
-emptyHomeModInfoLinkable = Nothing
-
-justBytecode :: Linkable -> Maybe Linkable
-justBytecode = Just
-
-justObjects :: Linkable -> Maybe Linkable
-justObjects = Just
-
-homeModInfoByteCode, homeModInfoObject :: HomeModInfo -> Maybe Linkable
-homeModInfoByteCode = hm_linkable
-homeModInfoObject = hm_linkable
-
-field_label :: a -> a
-field_label = id
-#endif
 
 mkSimpleTarget :: DynFlags -> FilePath -> Target
 mkSimpleTarget df fp = Target (TargetFile fp Nothing) True (homeUnitId_ df) Nothing
